import morepath

from onegov.core.security import Public, Private
from onegov.file import File
from onegov.org import _, OrgApp
from onegov.org.elements import Link
from onegov.org.forms import ImageSetForm
from onegov.org.layout import ImageSetLayout, ImageSetCollectionLayout
from onegov.org.models import (
    ImageFile,
    ImageSet,
    ImageSetCollection,
    ImageFileCollection
)
from purl import URL
from unidecode import unidecode


def get_form_class(self, request):
    if isinstance(self, ImageSetCollection):
        model = ImageSet()
    else:
        model = self

    return model.with_content_extensions(ImageSetForm, request)


@OrgApp.html(model=ImageSetCollection, template='imagesets.pt',
             permission=Public)
def view_imagesets(self, request):

    # XXX add collation support to the core (create collations automatically)
    imagesets = self.query().all()
    imagesets = sorted(imagesets, key=lambda d: unidecode(d.title))

    return {
        'layout': ImageSetCollectionLayout(self, request),
        'title': _("Photo Albums"),
        'imagesets': request.exclude_invisible(imagesets)
    }


@OrgApp.html(model=ImageSet, name='auswahl', template='select_images.pt',
             permission=Private, request_method='GET')
def select_images(self, request):

    collection = ImageFileCollection(request.app.session())
    selected = {f.id for f in self.files}

    def produce_image(id):
        return {
            'id': id,
            'src': request.class_link(File, {'id': id}, 'thumbnail'),
            'selected': id in selected
        }

    images = [
        {
            'group': request.translate(group),
            'images': tuple(produce_image(id) for group, id in items)
        } for group, items in collection.grouped_by_date()
    ]

    layout = ImageSetLayout(self, request)
    layout.breadcrumbs.append(Link(_("Select"), '#'))

    action = URL(request.link(self, 'auswahl')).query_param(
        'csrf-token', request.new_csrf_token())

    return {
        'layout': layout,
        'title': _("Select images"),
        'images': images,
        'action': action
    }


@OrgApp.html(model=ImageSet, name='auswahl', template='select_images.pt',
             permission=Private, request_method='POST')
def handle_select_images(self, request):

    # we do custom form handling here, so we need to check for CSRF manually
    request.assert_valid_csrf_token()

    if not request.POST:
        self.files = []
    else:
        self.files = request.app.session().query(ImageFile)\
            .filter(ImageFile.id.in_(request.POST)).all()

    request.success(_("Your changes were saved"))

    return morepath.redirect(request.link(self))


@OrgApp.form(model=ImageSetCollection, name='neu', template='form.pt',
             permission=Private, form=get_form_class)
def handle_new_imageset(self, request, form):

    if form.submitted(request):
        imageset = self.add(title=form.title.data)
        form.populate_obj(imageset)
        request.success(_("Added a new photo album"))

        return morepath.redirect(request.link(imageset))

    layout = ImageSetCollectionLayout(self, request)
    layout.include_editor()
    layout.breadcrumbs.append(Link(_("New"), '#'))

    return {
        'layout': layout,
        'title': _("New Photo Album"),
        'form': form,
    }


@OrgApp.form(model=ImageSet, name='bearbeiten', template='form.pt',
             permission=Private, form=get_form_class)
def handle_edit_imageset(self, request, form):
    if form.submitted(request):
        form.populate_obj(self)

        request.success(_("Your changes were saved"))
        return morepath.redirect(request.link(self))

    elif not request.POST:
        form.process(obj=self)

    layout = ImageSetLayout(self, request)
    layout.include_editor()
    layout.breadcrumbs.append(Link(_("Edit"), '#'))

    return {
        'layout': layout,
        'title': self.title,
        'form': form
    }


@OrgApp.view(model=ImageSet, request_method='DELETE', permission=Private)
def handle_delete_imageset(self, request):
    request.assert_valid_csrf_token()

    collection = ImageSetCollection(request.app.session())
    collection.delete(self)


@OrgApp.html(model=ImageSet, template='imageset.pt', permission=Public)
def view_imageset(self, request):

    return {
        'layout': ImageSetLayout(self, request),
        'title': self.title,
        'imageset': self
    }



from onegov.form import Form
from onegov.org import _
from onegov.user import UserCollection
from wtforms import HiddenField, StringField, PasswordField, validators


class RequestPasswordResetForm(Form):
    """ Defines the password reset request form for onegov org. """

    email = StringField(
        _("Email Address"),
        [validators.InputRequired(), validators.Email()]
    )

    def get_token(self, request):
        """ Returns the user and a token for the given username to reset the
        password. If the username is not found, (None, None) is returned.

        """
        users = UserCollection(request.app.session())
        user = users.by_username(self.email.data)
        token = None

        if user is not None:
            modified = user.modified.isoformat() if user.modified else ''
            token = request.new_url_safe_token({
                'username': user.username,
                'modified': modified
            })
        return user, token


class PasswordResetForm(Form):
    """ Defines the password reset form for onegov org. """

    email = StringField(
        _("Email Address"),
        [validators.InputRequired(), validators.Email()]
    )
    password = PasswordField(
        _("New Password"),
        [validators.InputRequired(), validators.Length(min=8)]
    )
    token = HiddenField()

    def update_password(self, request):
        """ Updates the password using the form data (if permitted to do so).

        Returns True if successful, False if not successful.
        """
        data = request.load_url_safe_token(self.token.data, max_age=86400)

        if not data:
            return False

        if not data.get('username') == self.email.data:
            return False

        users = UserCollection(request.app.session())
        user = users.by_username(self.email.data)

        if not user:
            return False

        modified = user.modified.isoformat() if user.modified else ''

        if modified != data['modified']:
            return False

        user.password = self.password.data

        return True



from datetime import datetime, timedelta
from onegov.core.templates import render_template
from onegov.org import _, OrgApp
from onegov.org.layout import DefaultMailLayout
from onegov.ticket import Ticket, TicketCollection
from onegov.user import User, UserCollection
from sedate import replace_timezone, to_timezone
from sqlalchemy.orm import undefer


MON = 0
SAT = 5
SUN = 6


@OrgApp.cronjob(hour=8, minute=30, timezone='Europe/Zurich')
def send_daily_ticket_statistics(request):

    today = replace_timezone(datetime.utcnow(), 'UTC')
    today = to_timezone(today, 'Europe/Zurich')

    if today.weekday() in (SAT, SUN):
        return

    args = {}

    # get the current ticket count
    collection = TicketCollection(request.app.session())
    count = collection.get_count()
    args['currently_open'] = count.open
    args['currently_pending'] = count.pending
    args['currently_closed'] = count.closed

    # get tickets created yesterday or on the weekend
    end = datetime(today.year, today.month, today.day, tzinfo=today.tzinfo)
    if today.weekday() == MON:
        start = end - timedelta(days=2)
    else:
        start = end - timedelta(days=1)

    query = collection.query()
    query = query.filter(Ticket.created >= start)
    query = query.filter(Ticket.created <= end)
    args['opened'] = query.count()

    query = collection.query()
    query = query.filter(Ticket.modified >= start)
    query = query.filter(Ticket.modified <= end)
    query = query.filter(Ticket.state == 'pending')
    args['pending'] = query.count()

    query = collection.query()
    query = query.filter(Ticket.modified >= start)
    query = query.filter(Ticket.modified <= end)
    query = query.filter(Ticket.state == 'closed')
    args['closed'] = query.count()

    # render the email
    args['title'] = request.translate(
        _("${org} OneGov Cloud Status", mapping={
            'org': request.app.org.name
        })
    )
    args['layout'] = DefaultMailLayout(object(), request)
    args['is_monday'] = today.weekday() == MON
    args['org'] = request.app.org.name

    # send one e-mail per user
    users = UserCollection(request.app.session()).query()
    users = users.filter(User.active == True)
    users = users.filter(User.role.in_(('admin', 'editor')))
    users = users.options(undefer('data'))
    users = users.all()

    for user in users:

        if user.data and not user.data.get('daily_ticket_statistics'):
            continue

        args['username'] = user.username
        content = render_template(
            'mail_daily_ticket_statistics.pt', request, args
        )

        request.app.send_email(
            subject=args['title'],
            receivers=(user.username, ),
            content=content
        )



from onegov.org import _, OrgApp
from onegov.org.elements import Link
from onegov.org.models import GeneralFileCollection, ImageFileCollection
from onegov.user.auth import Auth, UserCollection


@OrgApp.template_variables()
def get_template_variables(request):
    return {
        'bottom_links': tuple(get_bottom_links(request))
    }


def get_bottom_links(request):

    if request.is_logged_in:
        yield Link(_("Logout"), request.link(
            Auth.from_request_path(request), name='logout'))

        yield Link(_("User Profile"), request.link(
            request.app.org, name='benutzerprofil'))

    if request.is_manager:
        yield Link(_('Files'), request.class_link(
            GeneralFileCollection))

        yield Link(_('Images'), request.class_link(
            ImageFileCollection))

    if request.is_admin:
        yield Link(_('Settings'), request.link(
            request.app.org, 'einstellungen'))

        yield Link(_('Usermanagement'), request.class_link(
            UserCollection))

    if not request.is_logged_in:
        yield Link(_("Login"), request.link(
            Auth.from_request_path(request), name='login'))

        if request.app.settings.org.enable_user_registration:
            yield Link(_("Register"), request.link(
                Auth.from_request_path(request), name='register'
            ))

    yield Link('OneGov Cloud', 'http://www.onegovcloud.ch')
    yield Link('Seantis GmbH', 'https://www.seantis.ch')



import morepath

from elasticsearch import TransportError
from onegov.core.security import Public
from onegov.org import _, log, OrgApp
from onegov.org.elements import Link
from onegov.org.layout import DefaultLayout
from onegov.org.models import Search
from webob import exc


@OrgApp.html(model=Search, template='search.pt', permission=Public)
def search(self, request):

    layout = DefaultLayout(self, request)
    layout.breadcrumbs.append(Link(_("Search"), '#'))

    try:
        request.app.es_client.ping()
    except TransportError:
        log.warn("Elasticsearch cluster is offline")
        return {
            'title': _("Search Unavailable"),
            'layout': layout,
            'connection': False
        }

    if 'lucky' in request.GET:
        url = self.feeling_lucky()

        if url:
            return morepath.redirect(url)

    return {
        'title': _("Search"),
        'model': self,
        'layout': layout,
        'hide_search_header': True,
        'searchlabel': _("Search through ${count} indexed documents", mapping={
            'count': self.available_documents
        }),
        'resultslabel': _("${count} Results", mapping={
            'count': self.subset_count
        }),
        'connection': True
    }


@OrgApp.json(model=Search, name='suggest', permission=Public)
def suggestions(self, request):
    try:
        return tuple(self.suggestions())
    except TransportError:
        raise exc.HTTPNotFound()



from onegov.form import Form
from onegov.org import _
from wtforms import BooleanField


class UserProfileForm(Form):
    """ Defines the settings form for user profiles. """

    daily_ticket_statistics = BooleanField(_(
        "Send me a daily e-mail with information about the website's state."
    ))

    def update_model(self, model):
        if model.data is None:
            model.data = {}

        model.data['daily_ticket_statistics']\
            = self.data['daily_ticket_statistics']

    def apply_model(self, model):
        self.daily_ticket_statistics.data\
            = (model.data or {}).get('daily_ticket_statistics', True)



from collections import OrderedDict
from onegov.core.orm.mixins import meta_property, content_property
from onegov.core.utils import linkify
from onegov.form.parser.core import WTFormsClassBuilder, FieldDependency
from onegov.gis import CoordinatesField
from onegov.org import _
from onegov.people import Person, PersonCollection
from sqlalchemy.orm import object_session
from wtforms import BooleanField, StringField, TextAreaField


class ContentExtension(object):
    """ Extends classes based on :class:`onegov.core.orm.mixins.ContentMixin`
    with custom data that is stored in either 'meta' or 'content'.

    """

    @property
    def content_extensions(self):
        """ Returns all base classes of the current class which themselves have
        ``ContentExtension`` as baseclass.

        """
        for cls in self.__class__.__bases__:
            if ContentExtension in cls.__bases__:
                yield cls

    def with_content_extensions(self, form_class, request, extensions=None):
        """ Takes the given form and request and extends the form with
        all content extensions in the order in which they occur in the base
        class list.

        In other words, extends the forms with all used extension-fields.

        """

        for extension in extensions or self.content_extensions:
            form_class = extension.extend_form(self, form_class, request)

        return form_class

    def extend_form(self, form_class, request):
        """ Must be implemented by each ContentExtension. Takes the form
        class without extension and adds the required fields to it.

        """
        raise NotImplementedError


class HiddenFromPublicExtension(ContentExtension):
    """ Extends any class that has a meta dictionary field with the ability to
    hide it from the public.

    see :func:`onegov.core.security.rules.has_permission_not_logged_in`

    """

    is_hidden_from_public = meta_property('is_hidden_from_public')

    def extend_form(self, form_class, request):

        class HiddenPageForm(form_class):
            is_hidden_from_public = BooleanField(_("Hide from the public"))

        return HiddenPageForm


class CoordinatesExtension(ContentExtension):
    """ Extends any class that has a data dictionary field with the ability
    to add coordinates to it.

    """

    coordinates = content_property('coordinates')

    @property
    def has_coordinates(self):
        if self.coordinates:
            return self.coordinates.get('lat') and self.coordinates.get('lon')
        else:
            return False

    @property
    def lat(self):
        return self.coordinates.get('lat')

    @property
    def lon(self):
        return self.coordinates.get('lon')

    @property
    def zoom(self):
        return self.coordinates.get('zoom')

    def extend_form(self, form_class, request):

        class CoordinatesForm(form_class):
            coordinates = CoordinatesField(
                label=_("Coordinates"),
                description=_(
                    "The marker can be moved by dragging it with the mouse"
                ),
                fieldset=_("Map"),
                render_kw={'data-map-type': 'marker'}
            )

        return CoordinatesForm


class VisibleOnHomepageExtension(ContentExtension):
    """ Extends any class that has a meta dictionary field with the ability to
    a boolean indicating if the page should be shown on the homepage or not.

    """

    is_visible_on_homepage = meta_property('is_visible_on_homepage')

    def extend_form(self, form_class, request):

        # do not show on root pages
        if self.parent_id is None:
            return form_class

        class VisibleOnHomepageForm(form_class):
            is_visible_on_homepage = BooleanField(_("Visible on homepage"))

        return VisibleOnHomepageForm


class ContactExtension(ContentExtension):
    """ Extends any class that has a content dictionary field with a simple
    contacts field.

    """

    contact = content_property('contact')

    @contact.setter
    def contact(self, value):
        self.content['contact'] = value

        if value:
            self.content['contact_html'] = '<ul><li>{}</li></ul>'.format(
                '</li><li>'.join(linkify(value).splitlines())
            )
        else:
            self.content['contact_html'] = ""

    @property
    def contact_html(self):
        return self.content.get('contact_html')

    def extend_form(self, form_class, request):

        class ContactPageForm(form_class):
            contact = TextAreaField(
                label=_("Address"),
                fieldset=_("Contact"),
                render_kw={'rows': 5}
            )

        return ContactPageForm


class PersonLinkExtension(ContentExtension):
    """ Extends any class that has a content dictionary field with the ability
    to reference people from :class:`onegov.people.PersonCollection`.

    """

    @property
    def people(self):
        """ Returns the people linked to this content or None.

        The context specific function is temporarily stored on the
        ``context_specific_function`` attribute on each object in the
        resulting list.

        """

        if not self.content.get('people'):
            return None

        people = OrderedDict(self.content['people'])

        query = PersonCollection(object_session(self)).query()
        query = query.filter(Person.id.in_(people.keys()))

        result = []

        for person in query.all():
            person.context_specific_function = people[person.id.hex]
            result.append(person)

        order = list(people.keys())
        result.sort(key=lambda p: order.index(p.id.hex))

        return result

    def get_selectable_people(self, request):
        """ Returns a list of people which may be linked. """

        query = PersonCollection(request.app.session()).query()
        query = query.order_by(Person.last_name, Person.first_name)

        return query.all()

    def get_person_function_by_id(self, id):
        for _id, function in self.content.get('people', []):
            if id == _id:
                return function

    def move_person(self, subject, target, direction):
        """ Moves the subject below or above the target.

        :subject:
            The key of the person to be moved.

        :target:
            The key of the person above or below which the subject is moved.

        :direction:
            The direction relative to the target. Either 'above' or 'below'.

        """
        assert direction in ('above', 'below')
        assert subject != target
        assert self.content.get('people')

        def new_order():
            subject_function = self.get_person_function_by_id(subject)
            target_function = self.get_person_function_by_id(target)

            for person, function in self.content['people']:

                if person == subject:
                    continue

                if person == target and direction == 'above':
                    yield subject, subject_function
                    yield target, target_function
                    continue

                if person == target and direction == 'below':
                    yield target, target_function
                    yield subject, subject_function
                    continue

                yield person, function

        self.content['people'] = list(new_order())

    def extend_form(self, form_class, request):

        # XXX this is kind of implicitly set by the builder
        fieldset_id = 'people'
        fieldset_label = _("People")

        class PeoplePageForm(form_class):

            def get_people_fields(self, with_function):
                for field_id, field in self._fields.items():
                    if field_id.startswith(fieldset_id):
                        if with_function or not field_id.endswith('_function'):
                            yield field_id, field

            def get_people_and_function(self, selected_only=True):
                fields = self.get_people_fields(with_function=False)

                for field_id, field in fields:
                    if not selected_only or field.data is True:
                        person_id = field.id.hex
                        function = self._fields[field_id + '_function'].data

                        yield person_id, function

            def is_ordered_people(self, existing_people):
                """ Returns True if the current list of people is ordered
                from A to Z.

                """
                if not existing_people:
                    return True

                ordered_people = OrderedDict(self.get_people_and_function(
                    selected_only=False
                ))

                existing_people = [
                    key for key, value in existing_people
                    if key in ordered_people
                ]

                sorted_existing_people = sorted(
                    existing_people, key=list(ordered_people.keys()).index)

                return existing_people == sorted_existing_people

            def populate_obj(self, obj, *args, **kwargs):
                # XXX this no longer be necessary once the person links
                # have been turned into a field, see #74
                super().populate_obj(obj, *args, **kwargs)
                self.update_model(obj)

            def process(self, *args, **kwargs):
                # XXX this no longer be necessary once the person links
                # have been turned into a field, see #74
                super().process(*args, **kwargs)

                obj = kwargs.get('obj')

                if obj:
                    self.apply_model(obj)

            def update_model(self, model):
                previous_people = model.content.get('people', [])

                if self.is_ordered_people(previous_people):
                    # if the people are ordered a-z, we take the ordering from
                    # get_people_and_function, which comes by A-Z already
                    model.content['people'] = list(
                        self.get_people_and_function()
                    )
                else:
                    # if the people are not ordered we keep the order of the
                    # existing list and add the new people at the end
                    existing = set()
                    selected = {
                        key for key, function
                        in self.get_people_and_function()
                    }

                    old_people = list()
                    new_people = list()

                    for id, function in previous_people:
                        if id in selected:
                            old_people.append((id, function))
                            existing.add(id)

                    for id, function in self.get_people_and_function():
                        if id not in existing:
                            new_people.append((id, function))

                    model.content['people'] = old_people + new_people

            def apply_model(self, model):
                fields = self.get_people_fields(with_function=False)
                people = dict(model.content.get('people', []))

                for field_id, field in fields:
                    if field.id.hex in people:
                        self._fields[field_id].data = True
                        self._fields[field_id + '_function'].data\
                            = people[field.id.hex]

        builder = WTFormsClassBuilder(PeoplePageForm)
        builder.set_current_fieldset(fieldset_label)

        for person in self.get_selectable_people(request):
            field_id = builder.add_field(
                field_class=BooleanField,
                label=person.title,
                required=False,
                id=person.id
            )
            builder.add_field(
                field_class=StringField,
                label=_("Function"),
                required=False,
                dependency=FieldDependency(field_id, 'y')
            )

        return builder.form_class



from onegov.core.utils import sanitize_html
from onegov.form import Form
from onegov.form.validators import ValidFormDefinition
from onegov.org import _
from onegov.org.utils import annotate_html
from wtforms import StringField, TextAreaField, validators


class FormDefinitionBaseForm(Form):
    """ Form to edit defined forms. """

    title = StringField(_("Title"), [validators.InputRequired()])

    lead = TextAreaField(
        label=_("Lead"),
        description=_("Describes what this form is about"),
        render_kw={'rows': 4})

    text = TextAreaField(
        label=_("Text"),
        render_kw={'class_': 'editor'},
        filters=[sanitize_html, annotate_html])


class CustomDefinitionForm(FormDefinitionBaseForm):
    """ Same as the default form definition form, but with the definition
    made read-only.

    """

    definition = TextAreaField(
        label=_("Definition"),
        validators=[validators.InputRequired(), ValidFormDefinition()],
        render_kw={'rows': 32, 'data-editor': 'form'}
    )



import re

from lxml import etree
from onegov.core.utils import sanitize_html
from onegov.form import Form
from onegov.form.core import with_options
from onegov.gis import CoordinatesField
from onegov.org import _
from onegov.org.utils import annotate_html
from onegov.org.homepage_widgets import (
    transform_homepage_structure,
    XML_LINE_OFFSET
)
from onegov.org.theme import user_options
from wtforms import HiddenField, StringField, TextAreaField, validators
from wtforms.fields.html5 import EmailField, URLField
from wtforms_components import ColorField
from wtforms import ValidationError


ERROR_LINE_RE = re.compile(r'line ([0-9]+)')


class SettingsForm(Form):
    """ Defines the settings form for onegov org. """

    name = StringField(
        label=_("Name"),
        validators=[validators.InputRequired()],
        fieldset=_("General")
    )
    logo_url = StringField(
        label=_("Logo"),
        description=_("URL pointing to the logo"),
        fieldset=_("General"),
        render_kw={'class_': 'image-url'}
    )
    reply_to = EmailField(
        _("E-Mail Reply Address"), [validators.InputRequired()],
        description=_("Replies to automated e-mails go to this address."),
        fieldset=_("General")
    )
    primary_color = ColorField(
        label=_("Primary Color"),
        fieldset=_("General")
    )
    contact = TextAreaField(
        label=_("Contact"),
        description=_("The address and phone number of the municipality"),
        render_kw={'rows': 8},
        fieldset=_("Information")
    )
    contact_url = URLField(
        label=_("Contact Link"),
        description=_("URL pointing to a contact page"),
        fieldset=_("Information"),
        render_kw={'class_': 'internal-url'}
    )
    opening_hours = TextAreaField(
        label=_("Opening Hours"),
        description=_("The opening hours of the municipality"),
        render_kw={'rows': 8},
        fieldset=_("Information")
    )
    opening_hours_url = URLField(
        label=_("Opening Hours Link"),
        description=_("URL pointing to an opening hours page"),
        fieldset=_("Information"),
        render_kw={'class_': 'internal-url'}
    )
    facebook_url = URLField(
        label=_("Facebook"),
        description=_("URL pointing to the facebook site"),
        fieldset=_("Social Media")
    )
    twitter_url = URLField(
        label=_("Twitter"),
        description=_("URL pointing to the twitter site"),
        fieldset=_("Social Media")
    )
    homepage_image_1 = StringField(
        label=_("Homepage Image #1"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_image_2 = StringField(
        label=_("Homepage Image #2"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_image_3 = StringField(
        label=_("Homepage Image #3"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_image_4 = StringField(
        label=_("Homepage Image #4"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_image_5 = StringField(
        label=_("Homepage Image #5"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_image_6 = StringField(
        label=_("Homepage Image #6"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_cover = TextAreaField(
        label=_("Homepage Cover"),
        render_kw={'class_': 'editor', 'rows': 10},
        fieldset=_("Homepage"),
        filters=(sanitize_html, annotate_html)
    )
    homepage_structure = TextAreaField(
        label=_("Homepage Structure (for advanced users only)"),
        description=_("The structure of the homepage"),
        render_kw={'rows': 32, 'data-editor': 'xml'},
        fieldset=_("Homepage")
    )
    default_map_view = CoordinatesField(
        label=_("The default map view. This should show the whole town"),
        render_kw={
            'data-map-type': 'crosshair'
        },
        fieldset=_("Maps")
    )
    analytics_code = TextAreaField(
        label=_("Analytics Code"),
        description=_("JavaScript for web statistics support"),
        render_kw={'rows': 10},
        fieldset=_("Advanced")
    )

    # the footer height is determined by javascript, see org.scss and
    # common.js for more information (search for footer)
    footer_height = HiddenField()

    def validate_homepage_structure(self, field):
        if field.data:
            try:
                transform_homepage_structure(self.request.app, field.data)
            except etree.XMLSyntaxError as e:
                correct_line = e.position[0] - XML_LINE_OFFSET

                correct_msg = 'line {}'.format(correct_line)
                correct_msg = ERROR_LINE_RE.sub(correct_msg, e.msg)

                field.widget = with_options(
                    field.widget, **{'data-highlight-line': correct_line}
                )

                raise ValidationError(correct_msg)

    @property
    def theme_options(self):
        try:
            primary_color = self.primary_color.data.get_hex()
        except AttributeError:
            primary_color = user_options['primary-color']

        options = {
            'primary-color': primary_color,
            'footer-height': self.footer_height.data
        }

        # set the images only if provided
        for i in range(1, 7):
            image = getattr(self, 'homepage_image_{}'.format(i)).data

            if not image:
                continue

            options['tile-image-{}'.format(i)] = '"{}"'.format(image)

        # override the options using the default vaules if no value was given
        for key in options:
            if not options[key]:
                options[key] = user_options[key]

        return options

    @theme_options.setter
    def theme_options(self, options):
        self.primary_color.data = options.get('primary-color')
        self.footer_height.data = options.get('footer-height')
        self.homepage_image_1.data = options.get('tile-image-1', '').strip('"')
        self.homepage_image_2.data = options.get('tile-image-2', '').strip('"')
        self.homepage_image_3.data = options.get('tile-image-3', '').strip('"')
        self.homepage_image_4.data = options.get('tile-image-4', '').strip('"')
        self.homepage_image_5.data = options.get('tile-image-5', '').strip('"')
        self.homepage_image_6.data = options.get('tile-image-6', '').strip('"')



from cached_property import cached_property
from onegov.core.request import CoreRequest


class OrgRequest(CoreRequest):

    @cached_property
    def is_manager(self):
        """ Returns true if the current user is logged in, and has the role
        editor or admin.

        """

        return self.has_role('admin', 'editor')

    @cached_property
    def is_admin(self):
        """ Returns true if the current user is an admin.

        """

        return self.has_role('admin')



from onegov.core import utils
from onegov.core.orm.mixins import content_property
from onegov.page import Page
from onegov.search import ORMSearchable
from onegov.org.forms import LinkForm, PageForm
from onegov.org.models.atoz import AtoZ
from onegov.org.models.extensions import (
    ContactExtension,
    CoordinatesExtension,
    HiddenFromPublicExtension,
    PersonLinkExtension,
    VisibleOnHomepageExtension,
)
from onegov.org.models.traitinfo import TraitInfo
from sqlalchemy import desc, func
from sqlalchemy.orm import undefer, object_session


class SearchablePage(ORMSearchable):

    es_properties = {
        'title': {'type': 'localized'},
        'lead': {'type': 'localized'},
        'text': {'type': 'localized_html'}
    }

    lead = content_property('lead')
    text = content_property('text')
    url = content_property('url')

    @property
    def es_public(self):
        return not self.is_hidden_from_public

    @property
    def es_language(self):
        return 'de'  # xxx for now there's no other language

    @property
    def es_skip(self):
        return self.meta.get('trait') == 'link'  # do not index links

    @property
    def es_suggestions(self):
        return {
            "input": [self.title.lower()]
        }


class Topic(Page, TraitInfo, SearchablePage, HiddenFromPublicExtension,
            VisibleOnHomepageExtension, ContactExtension, PersonLinkExtension,
            CoordinatesExtension):
    __mapper_args__ = {'polymorphic_identity': 'topic'}

    es_type_name = 'topics'

    @property
    def deletable(self):
        """ Returns true if this page may be deleted. """
        return self.parent is not None

    @property
    def editable(self):
        return True

    @property
    def paste_target(self):
        if self.trait == 'link':
            return self.parent

        if self.trait == 'page':
            return self

        raise NotImplementedError

    @property
    def allowed_subtraits(self):
        if self.trait == 'link':
            return tuple()

        if self.trait == 'page':
            return ('page', 'link')

        raise NotImplementedError

    def is_supported_trait(self, trait):
        return trait in {'link', 'page'}

    def get_form_class(self, trait, request):
        if trait == 'link':
            return self.with_content_extensions(LinkForm, request, extensions=[
                HiddenFromPublicExtension,
                VisibleOnHomepageExtension
            ])

        if trait == 'page':
            return self.with_content_extensions(PageForm, request)

        raise NotImplementedError


class News(Page, TraitInfo, SearchablePage,
           HiddenFromPublicExtension, ContactExtension, PersonLinkExtension,
           CoordinatesExtension):
    __mapper_args__ = {'polymorphic_identity': 'news'}

    es_type_name = 'news'

    @property
    def absorb(self):
        return utils.lchop(self.path, 'aktuelles').lstrip('/')

    @property
    def deletable(self):
        return self.parent is not None

    @property
    def editable(self):
        return self.parent is not None

    @property
    def paste_target(self):
        if self.parent:
            return self.parent
        else:
            return self

    @property
    def allowed_subtraits(self):
        # only allow one level of news
        if self.parent is None:
            return ('news', )
        else:
            return tuple()

    def is_supported_trait(self, trait):
        return trait in {'news'}

    def get_form_class(self, trait, request):
        if trait == 'news':
            return self.with_content_extensions(PageForm, request)

        raise NotImplementedError

    @property
    def news_query(self):
        query = object_session(self).query(News)
        query = query.filter(Page.parent == self)
        query = query.order_by(desc(Page.created))
        query = query.options(undefer('created'))
        query = query.options(undefer('content'))

        return query

    @property
    def years(self):
        query = object_session(self).query(News)
        query = query.with_entities(func.date_part('year', Page.created))
        query = query.group_by(func.date_part('year', Page.created))
        query = query.filter(Page.parent == self)

        return sorted([int(r[0]) for r in query.all()], reverse=True)


class AtoZPages(AtoZ):

    def get_title(self, item):
        return item.title

    def get_items(self):

        # XXX implement correct collation support on the database level
        topics = self.request.app.session().query(Topic).all()
        topics = sorted(topics, key=self.sortkey)

        if self.request.is_manager:
            return [topic for topic in topics if topic.trait == 'page']
        else:
            return [
                topic for topic in topics if topic.trait == 'page' and
                not topic.is_hidden_from_public
            ]



import codecs
import os
import textwrap

from datetime import datetime, timedelta
from onegov.core.utils import module_path
from onegov.event import EventCollection
from onegov.form import FormCollection
from onegov.libres import ResourceCollection
from onegov.page import PageCollection
from onegov.org.models import Organisation
from sedate import as_datetime


def create_new_organisation(app, name):

    org = Organisation(name=name)
    org.homepage_structure = textwrap.dedent("""
        <row>
            <column span="8">
                <homepage-cover />
                <news />
            </column>
            <column span="4">
                <panel>
                    <links title="Dienstleistungen">
                        <link url="./formulare"
                            description="Anfragen &amp; Rückmeldungen">
                            Formulare
                        </link>
                        <link url="./ressourcen"
                            description="Räume &amp; Tageskarten">
                            Reservationen
                        </link>
                    </links>
                </panel>
                <panel>
                    <events />
                </panel>
                <panel>
                    <links title="Verzeichnisse">
                        <link url="./personen"
                            description="Alle Kontakte">
                            Personen
                        </link>
                        <link url="./fotoalben"
                            description="Impressionen">
                            Fotoalben
                        </link>
                        <link url="./a-z"
                            description="Kataolg A-Z">
                            Themen
                        </link>
                    </links>
                </panel>
            </column>
        </row>
    """)

    session = app.session()
    session.add(org)

    add_root_pages(session)
    add_builtin_forms(session)
    add_events(session, name)
    add_resources(app.libres_context)

    return org


def add_root_pages(session):
    pages = PageCollection(session)

    pages.add_root(
        "Organisation",
        name='organisation',
        type='topic',
        meta={'trait': 'page'}
    )
    pages.add_root(
        "Themen",
        name='themen',
        type='topic',
        meta={'trait': 'page'}
    )
    pages.add_root(
        "Kontakt",
        name='kontakt',
        type='topic',
        meta={'trait': 'page'}
    )
    pages.add_root(
        "Aktuelles",
        name='aktuelles',
        type='news',
        meta={'trait': 'news'}
    )


def add_builtin_forms(session, definitions=None):
    forms = FormCollection(session).definitions
    definitions = definitions or builtin_form_definitions()

    for name, title, definition in definitions:
        if not forms.by_name(name):
            form = forms.add(
                name=name,
                title=title,
                definition=definition,
                type='custom'
            )

        assert form.form_class().has_required_email_field, (
            "Each form must have at least one required email field"
        )


def builtin_form_definitions(path=None):
    """ Yields the name, title and the form definition of all form definitions
    in the given or the default path.

    """
    path = path or module_path('onegov.org', 'forms/builtin')

    for filename in os.listdir(path):
        if filename.endswith('.form'):
            name = filename.replace('.form', '')
            title, definition = load_definition(os.path.join(path, filename))
            yield name, title, definition


def load_definition(path):
    """ Loads the title and the form definition from the given file. """

    with codecs.open(path, 'r', encoding='utf-8') as formfile:
        formlines = formfile.readlines()

        title = formlines[0].strip()
        definition = ''.join(formlines[3:])

        return title, definition


def add_resources(libres_context):
    resource = ResourceCollection(libres_context)
    resource.add(
        "Tageskarte",
        'Europe/Zurich',
        type='daypass',
        name='tageskarte'
    )
    resource.add(
        "Konferenzraum",
        'Europe/Zurich',
        type='room',
        name='konferenzraum'
    )


def add_events(session, name):
    start = as_datetime(datetime.today().date())

    while start.weekday() != 6:
        start = start + timedelta(days=1)

    events = EventCollection(session)
    event = events.add(
        title="150 Jahre {}".format(name),
        start=start + timedelta(hours=11, minutes=0),
        end=start + timedelta(hours=22, minutes=0),
        timezone="Europe/Zurich",
        tags=["Party"],
        location="Sportanlage",
        content={
            "description": "Wir feiern das 150 jährige Bestehen.",
            "organizer": name
        },
        meta={"submitter_email": "info@example.org"},
    )
    event.submit()
    event.publish()
    event = events.add(
        title="Generalversammlung",
        start=start + timedelta(days=2, hours=20, minutes=0),
        end=start + timedelta(days=2, hours=22, minutes=30),
        timezone="Europe/Zurich",
        tags=["Politics"],
        location="Gemeindesaal",
        content={
            "description": "Alle Jahre wieder.",
            "organizer": name
        },
        meta={"submitter_email": "info@example.org"},
    )
    event.submit()
    event.publish()
    event = events.add(
        title="Gemeinsames Turnen",
        start=start + timedelta(days=2, hours=10, minutes=0),
        end=start + timedelta(days=2, hours=11, minutes=0),
        recurrence=(
            "RRULE:FREQ=WEEKLY;WKST=MO;BYDAY=TU,TH;UNTIL={0}".format(
                (start + timedelta(days=31)).strftime('%Y%m%dT%H%M%SZ')
            )
        ),
        timezone="Europe/Zurich",
        tags=["Sports"],
        location="Turnhalle",
        content={
            "description": "Gemeinsam fit werden.",
            "organizer": "Frauenverein"
        },
        meta={"submitter_email": "info@example.org"},
    )
    event.submit()
    event.publish()
    event = events.add(
        title="Grümpelturnier",
        start=start + timedelta(days=7, hours=10, minutes=0),
        end=start + timedelta(days=7, hours=18, minutes=0),
        timezone="Europe/Zurich",
        tags=["Sports"],
        location="Sportanlage",
        content={
            "description": "Bolzen auf dem Platz.",
            "organizer": "Sportverein"
        },
        meta={"submitter_email": "info@example.org"},
    )
    event.submit()
    event.publish()



from collections import namedtuple
from onegov.event import OccurrenceCollection
from onegov.newsletter import NewsletterCollection
from onegov.org import _, OrgApp
from onegov.org.elements import Link, LinkGroup
from onegov.org.layout import EventBaseLayout


@OrgApp.homepage_widget(tag='row')
class RowWidget(object):
    template = """
        <xsl:template match="row">
            <div class="row">
                <xsl:apply-templates select="node()"/>
            </div>
        </xsl:template>
    """


@OrgApp.homepage_widget(tag='column')
class ColumnWidget(object):
    template = """
        <xsl:template match="column">
            <div class="small-12 medium-{@span} columns">
                <xsl:apply-templates select="node()"/>
            </div>
        </xsl:template>
    """


@OrgApp.homepage_widget(tag='panel')
class PanelWidget(object):
    template = """
        <xsl:template match="panel">
            <div class="homepage-panel">
                <xsl:apply-templates select="node()"/>
            </div>
        </xsl:template>
    """


@OrgApp.homepage_widget(tag='links')
class LinksWidget(object):
    template = """
        <xsl:template match="links">
            <xsl:if test="@title">
                <h2>
                    <xsl:value-of select="@title" />
                </h2>
            </xsl:if>
            <ul class="panel-links">
                <xsl:for-each select="link">
                <li>
                    <a>
                        <xsl:attribute name="href">
                            <xsl:value-of select="@url" />
                        </xsl:attribute>

                        <xsl:value-of select="node()" />
                    </a>

                    <xsl:if test="@description">
                        <small>
                            <xsl:value-of select="@description" />
                        </small>
                    </xsl:if>
                </li>
                </xsl:for-each>
            </ul>
        </xsl:template>
    """


@OrgApp.homepage_widget(tag='news')
class NewsWidget(object):
    template = """
        <xsl:template match="news">
            <div metal:use-macro="layout.macros.newslist"
                tal:define="heading 'h3'; show_all_news_link True"
            />
        </xsl:template>
    """

    def get_variables(self, layout):
        return {
            'news': layout.root_pages[-1].news_query.limit(2).all()
        }


@OrgApp.homepage_widget(tag='homepage-cover')
class CoverWidget(object):
    template = """
        <xsl:template match="homepage-cover">
            <div class="homepage-content page-text">
                <tal:block
                    content="structure layout.org.meta.get('homepage_cover')"
                />
            </div>
        </xsl:template>
    """


@OrgApp.homepage_widget(tag='events')
class EventsWidget(object):
    template = """
        <xsl:template match="events">
            <metal:block use-macro="layout.macros['events-panel']" />
        </xsl:template>
    """

    def get_variables(self, layout):
        occurrences = OccurrenceCollection(layout.app.session()).query()
        occurrences = occurrences.limit(4)

        event_layout = EventBaseLayout(layout.model, layout.request)
        event_links = [
            Link(
                text=o.title,
                url=layout.request.link(o),
                subtitle=event_layout.format_date(o.localized_start, 'event')
            ) for o in occurrences
        ]

        event_links.append(
            Link(
                text=_("All events"),
                url=event_layout.events_url,
                classes=('more-link', )
            )
        )

        latest_events = LinkGroup(
            title=_("Events"),
            links=event_links,
        )

        return {
            'events_panel': latest_events
        }


@OrgApp.homepage_widget(tag='homepage-tiles')
class TilesWidget(object):
    template = """
        <xsl:template match="homepage-tiles">
            <xsl:choose>
                <xsl:when test="@show-title">
                    <metal:block use-macro="layout.macros['homepage-tiles']"
                       tal:define="show_title True" />
                </xsl:when>
                <xsl:otherwise>
                    <metal:block use-macro="layout.macros['homepage-tiles']"
                      tal:define="show_title False" />
                </xsl:otherwise>
            </xsl:choose>
        </xsl:template>
    """

    def get_variables(self, layout):
        return {'tiles': tuple(self.get_tiles(layout))}

    def get_tiles(self, layout):
        Tile = namedtuple('Tile', ['page', 'links', 'number'])

        homepage_pages = layout.request.app.homepage_pages
        request = layout.request
        link = request.link
        session = layout.app.session()
        classes = ('tile-sub-link', )

        for ix, page in enumerate(layout.root_pages):
            if page.type == 'topic':

                children = homepage_pages.get(page.id, tuple())
                children = (session.merge(c, load=False) for c in children)

                if not request.is_manager:
                    children = (
                        c for c in children if not c.is_hidden_from_public
                    )

                yield Tile(
                    page=Link(page.title, link(page)),
                    number=ix + 1,
                    links=tuple(
                        Link(c.title, link(c), classes=classes, model=c)
                        for c in children
                    )
                )
            elif page.type == 'news':
                news_url = link(page)
                years = (str(year) for year in page.years)

                links = [
                    Link(year, news_url + '?year=' + year, classes=classes)
                    for year in years
                ]

                links.append(Link(
                    _("Newsletter"), link(NewsletterCollection(session)),
                    classes=classes
                ))

                yield Tile(
                    page=Link(page.title, news_url),
                    number=ix + 1,
                    links=links
                )
            else:
                raise NotImplementedError



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from onegov.core.upgrade import upgrade_task
from onegov.org.models import Organisation


@upgrade_task('Move from town to organisation', always_run=True)
def move_town_to_organisation(context):

    # if the organisations table does not exist (other modules), skip
    has_organisations = context.app.session_manager.session().execute("""
        select exists (select 1 from information_schema.tables
        where table_schema='{}' and table_name='organisations')
    """.format(context.app.schema)).scalar()

    has_towns = context.app.session_manager.session().execute("""
        select exists (select 1 from information_schema.tables
        where table_schema='{}' and table_name='towns')
    """.format(context.app.schema)).scalar()

    if not has_organisations or not has_towns:
        return

    # this needs to be an always-run task for now, since on first install the
    # module is deteced as new and it is assumed that no tasks need to be run
    # as a result
    if context.session.query(Organisation).first():
        return

    context.operations.drop_table('organisations')
    context.operations.rename_table('towns', 'organisations')



import morepath

from collections import defaultdict
from onegov.core.crypto import random_password
from onegov.core.security import Secret
from onegov.org import _, OrgApp
from onegov.org.elements import Link
from onegov.org.forms import ManageUserForm, NewUserForm
from onegov.org.layout import UserManagementLayout
from onegov.user import User, UserCollection
from onegov.user.errors import ExistingUserError


@OrgApp.html(model=UserCollection, template='usermanagement.pt',
             permission=Secret)
def view_usermanagement(self, request):
    """ Allows the management of organisation users. """

    layout = UserManagementLayout(self, request)

    users = defaultdict(list)

    for user in self.query().order_by(User.username).all():
        users[user.role].append(user)

    return {
        'layout': layout,
        'title': _("User Management"),
        'users': users
    }


@OrgApp.form(model=User, template='form.pt', form=ManageUserForm,
             permission=Secret)
def handle_manage_user(self, request, form):

    # XXX the manage user form doesn't have access to the username
    # because it can't be edited, so we need to inject it here
    # for validation purposes (check for a unique yubikey)
    form.current_username = self.username

    if not request.app.settings.org.enable_yubikey:
        form.delete_field('yubikey')

    if form.submitted(request):
        form.populate_obj(self)
        request.success(_("Your changes were saved"))

        return morepath.redirect(request.class_link(UserCollection))
    else:
        form.process(obj=self)

    layout = UserManagementLayout(self, request)
    layout.breadcrumbs.append(Link(self.username, '#'))

    return {
        'layout': layout,
        'title': self.username,
        'form': form
    }


@OrgApp.form(model=UserCollection, template='newuser.pt', form=NewUserForm,
             name='neu', permission=Secret)
def handle_new_user(self, request, form):

    if not request.app.settings.org.enable_yubikey:
        form.delete_field('yubikey')

    layout = UserManagementLayout(self, request)
    layout.breadcrumbs.append(Link(_("New User"), '#'))
    layout.editbar_links = None

    if form.submitted(request):
        password = random_password()

        try:
            self.add(
                username=form.username.data,
                password=password,
                role=form.role.data,
                active=form.active.data
            )
        except ExistingUserError:
            form.username.errors.append(
                _("A user with this e-mail address already exists"))
        else:
            request.info(_("The user was created successfully"))

            return {
                'layout': layout,
                'title': _("New User"),
                'username': form.username.data,
                'password': password
            }

    return {
        'layout': layout,
        'title': _("New User"),
        'form': form
    }



from onegov.core.utils import sanitize_html
from onegov.form import Form
from onegov.org import _
from onegov.org.utils import annotate_html
from wtforms import StringField, TextAreaField, validators
from wtforms.fields.html5 import URLField


class PageBaseForm(Form):
    """ Defines the base form for all pages. """
    title = StringField(
        label=_("Title"),
        validators=[validators.InputRequired()],
        render_kw={'autofocus': ''}
    )


class LinkForm(PageBaseForm):
    """ Defines the form for pages with the 'link' trait. """
    url = URLField(
        label=_("URL"),
        validators=[validators.InputRequired()],
        render_kw={'class_': 'image-url file-url internal-url'}
    )


class PageForm(PageBaseForm):
    """ Defines the form for pages with the 'page' trait. """
    lead = TextAreaField(
        label=_("Lead"),
        description=_("Describes what this page is about"),
        render_kw={'rows': 4})

    text = TextAreaField(
        label=_("Text"),
        render_kw={'class_': 'editor'},
        filters=[sanitize_html, annotate_html])



""" Contains the model describing the organisation proper. """

from onegov.core.orm import Base
from onegov.core.orm.mixins import meta_property, TimestampMixin
from onegov.core.orm.types import JSON, UUID
from onegov.core.utils import linkify
from onegov.org.theme import user_options
from sqlalchemy import Column, Text
from uuid import uuid4


class Organisation(Base, TimestampMixin):
    """ Defines the basic information associated with an organisation.

    It is assumed that there's only one organisation record in the schema!
    """

    __tablename__ = 'organisations'

    #: the id of the organisation, an automatically generated uuid
    id = Column(UUID, nullable=False, primary_key=True, default=uuid4)

    #: the name of the organisation
    name = Column(Text, nullable=False)

    #: the logo of the organisation
    logo_url = Column(Text, nullable=True)

    #: the theme options of the organisation
    theme_options = Column(JSON, nullable=True, default=user_options.copy)

    #: additional data associated with the organisation
    meta = Column(JSON, nullable=True, default=dict)

    # meta bound values
    contact = meta_property('contact')
    contact_url = meta_property('contact_url')
    opening_hours = meta_property('opening_hours')
    opening_hours_url = meta_property('opening_hours_url')
    reply_to = meta_property('reply_to')
    facebook_url = meta_property('facebook_url')
    twitter_url = meta_property('twitter_url')
    analytics_code = meta_property('analytics_code')
    online_counter_label = meta_property('online_counter_label')
    reservations_label = meta_property('reservations_label')
    daypass_label = meta_property('daypass_label')
    default_map_view = meta_property('default_map_view')
    homepage_structure = meta_property('homepage_structure')
    homepage_cover = meta_property('homepage_cover')

    @contact.setter
    def contact(self, value):
        self.meta['contact'] = value
        self.meta['contact_html'] = linkify(value).replace('\n', '<br>')

    @opening_hours.setter
    def opening_hours(self, value):
        self.meta['opening_hours'] = value
        self.meta['opening_hours_html'] = linkify(value).replace('\n', '<br>')



import colorsys
import sedate

from datetime import datetime, time
from functools import lru_cache
from isodate import parse_date, parse_datetime
from itertools import groupby
from libres.modules import errors as libres_errors
from lxml.html import fragments_fromstring, tostring
from onegov.org import _
from onegov.org.elements import DeleteLink, Link
from onegov.ticket import TicketCollection
from operator import attrgetter
from purl import URL


def djb2_hash(text, size):
    """ Implementation of the djb2 hash, a simple hash function with a
    configurable table size.

    ** Do NOT use for cryptography! **

    """
    # arbitrary large prime number to initialize
    hash = 5381

    # hash(i) = hash(i-1) * 33 + str[i]
    for char in text:
        hash = ((hash << 5) + hash) + ord(char)

    # Output: integer between 0 and size-1 (inclusive)
    return hash % size


@lru_cache(maxsize=32)
def get_user_color(username):
    """ Gets a user color for each username which is used for the
    user-initials-* elements. Each username is mapped to a color.

    Since the colorspace is very limited there are lots of collisions.

    :returns: The user color in an css rgb string.

    """

    h = 100 / djb2_hash(username, 360)
    l = 0.9
    s = 0.5

    r, g, b = colorsys.hls_to_rgb(h, l, s)

    return '#{0:02x}{1:02x}{2:02x}'.format(
        int(round(r * 255)),
        int(round(g * 255)),
        int(round(b * 255))
    )


def format_time_range(start, end):
    return correct_time_range('{:%H:%M} - {:%H:%M}'.format(start, end))


def correct_time_range(string):
    if string.endswith('00:00'):
        return string[:-5] + '24:00'
    return string


def add_class_to_node(node, classname):
    """ Adds the given classname to the given lxml node's class list. """

    if 'class' in node.attrib:
        node.attrib['class'] += ' ' + classname
    else:
        node.attrib['class'] = classname


def annotate_html(html):
    """ Takes the given html and annotates the following elements for some
    advanced styling:

        * Every paragraph containing an img element will be marked with the
          `has-img` class.

        * If a link is found which points to a youtube or a vimeo video, the
          link itself as well as the surrounding paragraph is marked
          with the `has-video` class

    """

    if not html:
        return html

    fragments = fragments_fromstring(html)

    # we perform a root xpath lookup, which will result in all paragraphs
    # being looked at - so we don't need to loop over all elements (yah, it's
    # a bit weird)
    for element in fragments[:1]:

        # instead of failing, lxml will return strings instead of elements if
        # they can't be parsed.. so we have to inspect the objects
        if not hasattr(element, 'xpath'):
            return html

        for paragraph in element.xpath('//p[img]'):
            add_class_to_node(paragraph, 'has-img')

        condition = ' or '.join(
            'starts-with(@href, "{}")'.format(url) for url in {
                'https://www.youtube.com',
                'https://www.vimeo.com',
                'https://vimeo',
            }
        )

        for a in element.xpath('//a[{}]'.format(condition)):
            add_class_to_node(a, 'has-video')

            # find the closest paragraph in the ancestrage, but don't look far
            parent = a.getparent()

            for i in range(0, 3):
                if parent.tag == 'p':
                    add_class_to_node(parent, 'has-video')
                    break
                parent = parent.getparent()

        for img in element.xpath('//img'):
            img.set('class', 'lazyload-alt')

    return ''.join(tostring(e).decode('utf-8') for e in fragments)


def parse_fullcalendar_request(request, timezone):
    """ Parses start and end from the given fullcalendar request. It is
    expected that no timezone is passed (the default).

    See `<http://fullcalendar.io/docs/timezone/timezone/>`_

    :returns: A tuple of timezone-aware datetime objects or (None, None).

    """
    start = request.params.get('start')
    end = request.params.get('end')

    if start and end:
        if 'T' in start:
            start = parse_datetime(start)
            end = parse_datetime(end)
        else:
            start = datetime.combine(parse_date(start), time(0, 0))
            end = datetime.combine(parse_date(end), time(23, 59, 59, 999999))

        start = sedate.replace_timezone(start, timezone)
        end = sedate.replace_timezone(end, timezone)

        return start, end
    else:
        return None, None


def render_time_range(start, end):
    start = '{:%H:%M}'.format(start)

    if (end.time() == time(0, 0)):
        end = '24:00'
    else:
        end = '{:%H:%M}'.format(end)

    return ' - '.join((start, end))


class ReservationInfo(object):

    __slots__ = ['resource', 'reservation', 'request', 'translate']

    def __init__(self, reservation, request):
        self.reservation = reservation
        self.request = request
        self.translate = request.translate

    @property
    def date(self):
        return self.reservation.display_start().isoformat()

    @property
    def time(self):
        if sedate.is_whole_day(
            self.reservation.start,
            self.reservation.end,
            self.reservation.timezone
        ):
            return self.translate(_("Whole day"))
        else:
            return render_time_range(
                self.reservation.display_start(),
                self.reservation.display_end()
            )

    @property
    def delete_link(self):
        url = self.request.link(self.reservation)
        url = URL(url).query_param('csrf-token', self.request.new_csrf_token())

        return url.as_string()

    def as_dict(self):
        return {
            'date': self.date,
            'time': self.time,
            'delete': self.delete_link,
            'quota': self.reservation.quota,
            'created': self.reservation.created.isoformat()
        }


class AllocationEventInfo(object):

    __slots__ = ['allocation', 'availability', 'request', 'translate']

    def __init__(self, allocation, availability, request):
        self.allocation = allocation
        self.availability = availability
        self.request = request
        self.translate = request.translate

    @classmethod
    def from_allocations(cls, request, scheduler, allocations):
        events = []

        for key, group in groupby(allocations, key=attrgetter('_start')):
            grouped = tuple(group)
            availability = scheduler.queries.availability_by_allocations(
                grouped
            )

            for allocation in grouped:
                if allocation.is_master:
                    events.append(
                        cls(
                            allocation,
                            availability,
                            request
                        )
                    )

        return events

    @property
    def event_start(self):
        return self.allocation.display_start().isoformat()

    @property
    def event_end(self):
        return self.allocation.display_end().isoformat()

    @property
    def event_identification(self):
        return '{:%d.%m.%Y}: {}'.format(
            self.allocation.display_start(),
            self.event_time
        )

    @property
    def event_time(self):
        if self.allocation.whole_day:
            return self.translate(_("Whole day"))
        else:
            return render_time_range(
                self.allocation.display_start(),
                self.allocation.display_end()
            )

    @property
    def quota(self):
        return self.allocation.quota

    @property
    def quota_left(self):
        return int(self.quota * self.availability / 100)

    @property
    def event_title(self):
        if self.allocation.partly_available:
            available = self.translate(_("${percent}% Available", mapping={
                'percent': int(self.availability)
            }))
        else:

            quota = self.quota
            quota_left = self.quota_left

            if quota == 1:
                if quota_left:
                    available = self.translate(_("Available"))
                else:
                    available = self.translate(_("Unavailable"))
            else:
                available = self.translate(
                    _("${num}/${max} Available", mapping={
                        'num': quota_left,
                        'max': quota
                    })
                )

        # add an extra space at the end of the event time, so we can hide
        # the <br> tag on the output without having the time and the
        # availability seemingly joined together without space.
        return '\n'.join((self.event_time + ' ', available))

    @property
    def event_class(self):
        if self.availability >= 80.0:
            return 'event-available'
        if self.availability >= 20.0:
            return 'event-partly-available'
        else:
            return 'event-unavailable'

    @property
    def event_actions(self):
        if self.request.is_manager:
            yield Link(
                _("Edit"),
                self.request.link(self.allocation, name='bearbeiten'),
            )

            yield Link(
                _("Tickets"),
                self.request.link(TicketCollection(
                    session=self.request.app.session,
                    handler='RSV',
                    state='all',
                    extra_parameters={
                        'allocation_id': str(self.allocation.id)
                    }
                )),
            )

            if self.availability == 100.0:
                yield DeleteLink(
                    _("Delete"),
                    self.request.link(self.allocation),
                    confirm=_("Do you really want to delete this allocation?"),
                    extra_information=self.event_identification,
                    yes_button_text=_("Delete allocation")
                )
            else:
                yield DeleteLink(
                    _("Delete"),
                    self.request.link(self.allocation),
                    confirm=_(
                        "This allocation can't be deleted because there are "
                        "existing reservations associated with it."
                    ),
                    extra_information=_(
                        "To delete this allocation, all existing reservations "
                        "need to be cancelled first."
                    )
                )

    def as_dict(self):
        return {
            'id': self.allocation.id,
            'start': self.event_start,
            'end': self.event_end,
            'title': self.event_title,
            'wholeDay': self.allocation.whole_day,
            'partlyAvailable': self.allocation.partly_available,
            'quota': self.allocation.quota,
            'quotaLeft': self.quota_left,
            'className': self.event_class,
            'partitions': self.allocation.availability_partitions(),
            'actions': [
                link(self.request).decode('utf-8')
                for link in self.event_actions
            ],
            'editurl': self.request.link(self.allocation, name='bearbeiten'),
            'reserveurl': self.request.link(self.allocation, name='reserve')
        }


libres_error_messages = {
    libres_errors.OverlappingAllocationError:
    _("A conflicting allocation exists for the requested time period."),

    libres_errors.OverlappingReservationError:
    _("A conflicting reservation exists for the requested time period."),

    libres_errors.AffectedReservationError:
    _("An existing reservation would be affected by the requested change."),

    libres_errors.AffectedPendingReservationError:
    _("A pending reservation would be affected by the requested change."),

    libres_errors.AlreadyReservedError:
    _("The requested period is no longer available."),

    libres_errors.NotReservableError:
    _("No reservable slot found."),

    libres_errors.ReservationTooLong:
    _("Reservations can't be made for more than 24 hours at a time."),

    libres_errors.ReservationParametersInvalid:
    _("The given reservation paramters are invalid."),

    libres_errors.InvalidReservationToken:
    _("The given reservation token is invalid."),

    libres_errors.InvalidReservationError:
    _("The given reservation paramters are invalid."),

    libres_errors.QuotaOverLimit:
    _("The requested number of reservations is higher than allowed."),

    libres_errors.InvalidQuota:
    _("The requested quota is invalid (must be at least one)."),

    libres_errors.QuotaImpossible:
    _("The allocation does not have enough free spots."),

    libres_errors.InvalidAllocationError:
    _("The resulting allocation would be invalid."),

    libres_errors.NoReservationsToConfirm:
    _("No reservations to confirm."),

    libres_errors.TimerangeTooLong:
    _("The given timerange is longer than the existing allocation."),

    libres_errors.ReservationTooShort:
    _("Reservation too short. A reservation must last at least 5 minutes.")
}


def get_libres_error(e, request):
    assert type(e) in libres_error_messages, (
        "Unknown libres error {}".format(type(e))
    )

    return request.translate(libres_error_messages.get(type(e)))


def show_libres_error(e, request):
    """ Shows a human readable error message for the given libres exception,
    using request.alert.

    """
    request.alert(get_libres_error(e, request))



import transaction

from onegov.core.utils import Bunch
from onegov.org import OrgApp


def test_allowed_application_id(org_app):
    # little bobby tables!
    assert not org_app.is_allowed_application_id(
        "Robert'); DROP TABLE Students; --"
    )
    assert not org_app.is_allowed_application_id('foo')
    org_app.session_manager.ensure_schema_exists(org_app.namespace + '-foo')
    assert org_app.is_allowed_application_id('foo')


def test_send_email(smtp):

    class App(OrgApp):

        @property
        def org(self):
            return Bunch(
                name="Gemeinde Govikon",
                meta=dict(reply_to='info@govikon.ch')
            )

    app = App()
    app.mail_host, app.mail_port = smtp.address
    app.mail_sender = 'mails@govikon.ch'
    app.mail_force_tls = False
    app.mail_username = None
    app.mail_password = None
    app.mail_use_directory = False

    app.send_email(receivers=['civilian@example.org'], subject='Test')
    assert len(smtp.outbox) == 0
    transaction.abort()

    app.send_email(receivers=['civilian@example.org'], subject='Test')
    assert len(smtp.outbox) == 0
    transaction.commit()

    assert len(smtp.outbox) == 1
    mail = smtp.outbox[0]

    assert mail['Reply-To'] == 'Gemeinde Govikon <info@govikon.ch>'
    assert mail['Subject'] == 'Test'
    assert mail['Sender'] == 'Gemeinde Govikon <mails@govikon.ch>'
    assert mail['From'] == 'Gemeinde Govikon <mails@govikon.ch>'



import logging
log = logging.getLogger('onegov.org')  # noqa
log.addHandler(logging.NullHandler())  # noqa

from translationstring import TranslationStringFactory  # noqa
_ = TranslationStringFactory('onegov.org')  # noqa

from onegov.org.app import OrgApp

# directives need to be imported to be captured by morepath
# unless until https://github.com/morepath/dectate/issues/37 is solved
from onegov.org.directive import *  # noqa

__all__ = ['OrgApp']



from onegov.org.models.atoz import AtoZ
from onegov.org.models.clipboard import Clipboard
from onegov.org.models.editor import Editor
from onegov.org.models.file import (
    GeneralFile,
    GeneralFileCollection,
    ImageFile,
    ImageFileCollection,
    ImageSet,
    ImageSetCollection,
)
from onegov.org.models.form import BuiltinFormDefinition, CustomFormDefinition
from onegov.org.models.legacy_file import (
    LegacyFile,
    LegacyImage,
    LegacyFileCollection,
    LegacyImageCollection,
)
from onegov.org.models.organisation import Organisation
from onegov.org.models.page import AtoZPages, News, Topic
from onegov.org.models.page_move import PageMove
from onegov.org.models.person_move import (
    FormPersonMove,
    PagePersonMove,
    PersonMove,
    ResourcePersonMove,
)
from onegov.org.models.resource import DaypassResource
from onegov.org.models.search import Search
from onegov.org.models.sitecollection import SiteCollection
from onegov.org.models.traitinfo import TraitInfo
from onegov.org.models.extensions import (
    ContactExtension,
    ContentExtension,
    CoordinatesExtension,
    HiddenFromPublicExtension,
    PersonLinkExtension,
    VisibleOnHomepageExtension,
)

__all__ = [
    'AtoZ',
    'AtoZPages',
    'BuiltinFormDefinition',
    'Clipboard',
    'ContactExtension',
    'ContentExtension',
    'CoordinatesExtension',
    'CustomFormDefinition',
    'DaypassResource',
    'Editor',
    'FormPersonMove',
    'GeneralFile',
    'GeneralFileCollection',
    'HiddenFromPublicExtension',
    'ImageFile',
    'ImageFileCollection',
    'ImageSet',
    'ImageSetCollection',
    'LegacyFile',
    'LegacyFileCollection',
    'LegacyImage',
    'LegacyImageCollection',
    'News',
    'Organisation',
    'PageMove',
    'PagePersonMove',
    'PersonLinkExtension',
    'PersonMove',
    'ResourcePersonMove',
    'Search',
    'SiteCollection',
    'Topic',
    'TraitInfo',
    'VisibleOnHomepageExtension',
]



from datetime import date, datetime, timedelta
from dateutil import rrule
from onegov.form import Form
from onegov.form.fields import MultiCheckboxField
from onegov.gis import CoordinatesField
from onegov.org import _
from sedate import replace_timezone, to_timezone
from wtforms import StringField, TextAreaField, validators
from wtforms.fields.html5 import DateField, EmailField
from wtforms_components import TimeField


TAGS = [(tag, tag) for tag in (
    _("Exhibition"),
    _("Gastronomy"),
    _("Health"),
    _("Cinema"),
    _("Congress"),
    _("Concert"),
    _("Art"),
    _("Literature"),
    _("Market"),
    _("Music"),
    _("Party"),
    _("Politics"),
    _("Religion"),
    _("Sports"),
    _("Dancing"),
    _("Theater"),
    _("Meetup"),
    _("Talk"),
)]

WEEKDAYS = (
    ("MO", _("Mo")),
    ("TU", _("Tu")),
    ("WE", _("We")),
    ("TH", _("Th")),
    ("FR", _("Fr")),
    ("SA", _("Sa")),
    ("SU", _("Su")),
)


class EventForm(Form):
    """ Defines the form for all events. """

    email = EmailField(
        label=_("E-Mail"),
        description="max.muster@example.org",
        validators=[validators.InputRequired(), validators.Email()]
    )

    title = StringField(
        label=_("Title"),
        description=_("Concerto in the castle garden"),
        validators=[validators.InputRequired()]
    )

    description = TextAreaField(
        label=_("Description"),
        description=_("Enjoy a concerto in the castle garden."),
        render_kw={'rows': 12}
    )

    location = StringField(
        label=_("Location"),
        description=_("Castle garden"),
        validators=[validators.InputRequired()]
    )

    organizer = StringField(
        label=_("Organizer"),
        description=_("Music society"),
        validators=[validators.InputRequired()]
    )

    coordinates = CoordinatesField(
        label=_("Coordinates"),
        description=_("The marker can be moved by dragging it with the mouse"),
        render_kw={'data-map-type': 'marker'}
    )

    tags = MultiCheckboxField(
        label=_("Tags"),
        choices=TAGS,
    )

    start_date = DateField(
        label=_("Date"),
        validators=[validators.InputRequired()],
        default=date.today
    )

    start_time = TimeField(
        label=_("From"),
        description="18:00",
        validators=[validators.InputRequired()]
    )

    end_time = TimeField(
        label=_("To"),
        description="19:15",
        validators=[validators.InputRequired()]
    )

    weekly = MultiCheckboxField(
        label=_("Repeats itself every"),
        choices=WEEKDAYS,
        render_kw={'prefix_label': False, 'class_': 'oneline-checkboxes'}
    )

    end_date = DateField(
        label=_("Until date"),
        validators=[validators.Optional()]
    )

    def validate(self):
        """ Make sure a valid RRULE can be generated with the given fields.

        Might be better to group weekly and end_date in an enclosure,
        see See http://wtforms.readthedocs.org/en/latest/fields.html
        #field-enclosures.

        """
        result = super().validate()

        if self.start_date.data and self.end_date.data:
            if self.start_date.data > self.end_date.data:
                message = _("The end date must be later than the start date.")
                self.end_date.errors.append(message)
                result = False

        if self.weekly.data and self.start_date.data:
            weekday = WEEKDAYS[self.start_date.data.weekday()][0]
            if weekday not in self.weekly.data:
                message = _("The weekday of the start date must be selected.")
                self.weekly.errors.append(message)
                result = False

        if self.weekly.data and not self.end_date.data:
            message = _("Please set and end date if the event is recurring.")
            self.end_date.errors.append(message)
            result = False

        if self.end_date.data and not self.weekly.data:
            message = _("Please select a weekday if the event is recurring.")
            self.weekly.errors.append(message)
            result = False

        return result

    def populate_obj(self, model):
        """ Stores the form values on the model. """

        super().populate_obj(model, exclude={
            'start_date',
            'start_time',
            'end_date',
            'end_time',
            'weekly',
            'email',
        })

        # clear the recurrence to avoid updating all occurrences too much
        model.recurrence = ''

        model.meta = model.meta or {}
        model.meta['submitter_email'] = self.email.data
        model.timezone = 'Europe/Zurich'

        model.start = replace_timezone(
            datetime(
                self.start_date.data.year,
                self.start_date.data.month,
                self.start_date.data.day,
                self.start_time.data.hour,
                self.start_time.data.minute
            ),
            model.timezone
        )
        end_date = self.start_date.data
        if self.end_time.data <= self.start_time.data:
            end_date += timedelta(days=1)

        model.end = replace_timezone(
            datetime(
                end_date.year,
                end_date.month,
                end_date.day,
                self.end_time.data.hour,
                self.end_time.data.minute
            ),
            model.timezone
        )

        if self.weekly.data and self.end_date.data:
            until_date = to_timezone(
                replace_timezone(
                    datetime(
                        self.end_date.data.year,
                        self.end_date.data.month,
                        self.end_date.data.day,
                        self.end_time.data.hour,
                        self.end_time.data.minute
                    ), model.timezone
                ), 'UTC'
            )
            model.recurrence = (
                "RRULE:FREQ=WEEKLY;WKST=MO;BYDAY={0};UNTIL={1}".format(
                    ','.join(self.weekly.data),
                    until_date.strftime('%Y%m%dT%H%M%SZ')
                )
            )

    def process(self, *args, **kwargs):
        """ Stores the page values on the form. """

        super().process(*args, **kwargs)

        if 'obj' in kwargs:
            model = kwargs['obj']

            self.start_time.data = model.localized_start.time()
            self.end_time.data = model.localized_end.time()
            self.start_date.data = model.localized_start.date()

            if model.recurrence:
                last_occurrence = model.occurrence_dates(localize=True)[-1]
                self.end_date.data = last_occurrence.date()

                rule = rrule.rrulestr(model.recurrence)
                if rule._freq == rrule.WEEKLY:
                    self.weekly.data = [
                        WEEKDAYS[day][0] for day in rule._byweekday
                    ]

            if model.meta:
                self.email.data = model.meta.get('submitter_email')



""" The onegov org collection of files uploaded to the site. """

import morepath

from onegov.core.filestorage import view_filestorage_file
from onegov.core.security import Private, Public
from onegov.file import File, FileCollection
from onegov.file.utils import IMAGE_MIME_TYPES
from onegov.org import _, OrgApp
from onegov.org.elements import Img, Link
from onegov.org.layout import DefaultLayout
from onegov.org.models import (
    GeneralFile,
    GeneralFileCollection,
    ImageFile,
    ImageFileCollection,
    ImageSetCollection,
    LegacyFile,
    LegacyImage,
)
from sedate import utcnow
from webob import exc


@OrgApp.html(model=GeneralFileCollection, template='files.pt',
             permission=Private)
def view_get_file_collection(self, request):
    request.include('dropzone')

    files = [
        Link(text=f.name, url=request.link(f))
        for f in self.query().order_by(File.name).all()
    ]

    layout = DefaultLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Files"), request.link(self))
    ]

    return {
        'layout': layout,
        'title': _('Files'),
        'files': files,
    }


@OrgApp.html(model=ImageFileCollection, template='images.pt',
             permission=Private)
def view_get_image_collection(self, request):
    request.include('common')
    request.include('dropzone')
    request.include('editalttext')

    layout = DefaultLayout(self, request)

    images = view_get_image_collection_json(
        self, request, produce_image=lambda image: Img(
            src=request.class_link(File, {'id': image.id}, 'thumbnail'),
            url=request.class_link(File, {'id': image.id}),
            alt=(image.note or '').strip(),
            extra=layout.csrf_protected_url(request.link(image, 'note'))
        )
    )

    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Images"), request.link(self))
    ]

    layout.editbar_links = [
        Link(
            text=_("Manage Photo Albums"),
            url=request.class_link(ImageSetCollection),
            classes=('new-photo-album', ))
    ]

    return {
        'layout': layout,
        'title': _('Images'),
        'images': images
    }


@OrgApp.json(model=GeneralFileCollection, permission=Private, name='json')
def view_get_file_collection_json(self, request):
    return [
        {
            'link': request.class_link(File, {'id': id}),
            'title': name
        }
        for id, name in self.query().with_entities(File.id, File.name).all()
    ]


@OrgApp.json(model=ImageFileCollection, permission=Private, name='json')
def view_get_image_collection_json(self, request, produce_image=None):

    if not produce_image:
        def produce_image(image):
            return {
                'thumb': request.class_link(
                    File, {'id': image.id}, 'thumbnail'),
                'image': request.class_link(
                    File, {'id': image.id})
            }

    return [
        {
            'group': request.translate(group),
            'images': tuple(produce_image(image) for group, image in items)
        } for group, items in self.grouped_by_date(id_only=False)
    ]


def handle_file_upload(self, request):
    """ Stores the file given with the request and returns the new file object.

    """

    return self.add(
        filename=request.params['file'].filename,
        content=request.params['file'].file
    )


@OrgApp.view(model=FileCollection, name='upload',
             request_method='POST', permission=Private)
def view_upload_file(self, request):
    request.assert_valid_csrf_token()

    try:
        file = handle_file_upload(self, request)
    except FileExistsError as e:
        # mark existing files as modified to put them in front of the queue
        e.args[0].modified = utcnow()
    else:
        if isinstance(self, GeneralFileCollection):
            pass
        elif isinstance(self, ImageFileCollection):
            if file.reference.content_type not in IMAGE_MIME_TYPES:
                raise exc.HTTPUnsupportedMediaType()
        else:
            raise NotImplementedError

    return morepath.redirect(request.link(self))


@OrgApp.json(model=FileCollection, name='upload.json',
             request_method='POST', permission=Private)
def view_upload_file_by_json(self, request):
    request.assert_valid_csrf_token()

    try:
        f = handle_file_upload(self, request)
    except FileExistsError as e:
        # mark existing files as modified to put them in front of the queue
        e.args[0].modified = utcnow()

        return {
            'filelink': request.link(e.args[0]),
            'filename': e.args[0].name
        }
    except exc.HTTPUnsupportedMediaType:
        return {
            'error': True,
            'message': request.translate(_("This file type is not supported"))
        }
    except exc.HTTPRequestHeaderFieldsTooLarge:
        return {
            'error': True,
            'message': request.translate(_("The file name is too long"))
        }

    return {
        'filelink': request.link(f),
        'filename': f.name,
    }


@OrgApp.view(model=LegacyFile, permission=Public)
@OrgApp.view(model=LegacyImage, permission=Public)
def view_old_files_redirect(self, request):
    """ Redirects to the migrated depot file if possible. As a result, old
    image urls are preserved and will continue to function.

    """

    alternate_path = self.path + '.r'

    if request.app.filestorage.isfile(alternate_path):
        with request.app.filestorage.open(alternate_path, 'r') as f:
            id = f.read()

        if isinstance(self, LegacyFile):
            file_class = GeneralFile
        else:
            file_class = ImageFile

        return exc.HTTPMovedPermanently(
            location=request.class_link(file_class, {'id': id}))

    return view_filestorage_file(self, request)
